use std::fs::File;
use std::io::prelude::*;
use std::path::{Path, PathBuf};
use std::time::Instant;

use anyhow::{bail, ensure};
use clap::Parser;

use enemize::{asar, rom::RomData};

const ASAR_SYMBOLS: &'static str = "asar_symbols.txt";
const ENEMIZER_BASE_PATCH: &'static str = "enemizer_base_patch.json";

/// Randomizes enemy placements in The Legend of Zelda: A Link to the Past for the Super Nintendo
/// Entertainment System
#[derive(Debug, Parser)]
#[clap(author, version, about)]
struct Args {
    /// path to the base rom file
    #[clap(short, long)]
    rom: PathBuf,
    /// path to the enemizerOptions.json
    #[clap(short, long)]
    enemizer: PathBuf,
    /// seed number
    #[clap(short, long)]
    seed: Option<i32>,
    /// path to the intended output file
    #[clap(short, long)]
    output: PathBuf,
    /// operate in binary mode (takes already randomized SFC and applies enemizer directly to ROM)
    #[clap(long)]
    binary: bool,
    /// path to base2patched.json (not used)
    #[clap(long)]
    base: Option<PathBuf>,
    /// path to the randomizerPatch.json (not used)
    #[clap(long)]
    randomizer: Option<PathBuf>,
}

fn main() -> anyhow::Result<()> {
    let args = Args::parse();

    let stopwatch = Instant::now();

    let options = {
        let mut opts_file = File::open(args.enemizer)?;
        let mut json = String::new();
        opts_file.read_to_string(&mut json)?;
        serde_json::from_str(&json)?
    };
    let symbols = load_symbols()?;

    let mut raw_data = vec![];
    let mut rom_file = File::open(args.rom)?;
    rom_file.read_to_end(&mut raw_data)?;

    raw_data.resize(2 * 1024 * 1024, 0);

    let mut rom = RomData::new(symbols, raw_data);
    ensure!(!rom.is_enemizer(), "It appears that the provided base ROM is already enemized. Please ensure you are using an original game ROM.");

    // Oh noes! The max seed number is twice as likely to show up as any other!
    // (That is, 2 out of 2 billion instead of 1.)
    let seed = args.seed.unwrap_or_else(|| rand::random()).saturating_abs();

    rom.randomize(Path::new(ENEMIZER_BASE_PATCH), options, seed)?;

    let mut out_file = File::create(&args.output)?;

    if args.binary {
        out_file.write_all(rom.get_rom_bytes())?;
        out_file.flush()?;
        println!("Generated SFC file {}", args.output.to_string_lossy());
    } else {
        let patches = rom.generate_patch();
        serde_json::to_writer(out_file, &patches)?;
    }

    println!("Seed generated in: {}ms", stopwatch.elapsed().as_millis());

    Ok(())
}

fn load_symbols() -> anyhow::Result<asar::Symbols> {
    if let Err(_) = std::fs::metadata(ASAR_SYMBOLS) {
        bail!(
            "Could not find symbols at {}. Did you run prepare.sh?",
            ASAR_SYMBOLS
        );
    }

    asar::load_symbols(Path::new(ASAR_SYMBOLS))
}
